@@ -25,7 +25,7 @@ class EdgeTransformer:
         self,
         method: str = "Hadamard",
         aligned_node_mapping: bool = False,
-        support_mirror_strategy: bool = False,
+        support_mirrored_strategy: bool = False,
     ):
         """Create new EdgeTransformer object.
 
@@ -40,7 +40,7 @@ class EdgeTransformer:
             matches the internal node mapping of the given graph.
             If these two mappings do not match, the generated edge embedding
             will be meaningless.
-        support_mirror_strategy: bool = False,
+        support_mirrored_strategy: bool = False,
             Wethever to patch support for mirror strategy.
             At the time of writing, TensorFlow's MirrorStrategy does not support
             input values different from floats, therefore to support it we need
@@ -59,7 +59,7 @@ class EdgeTransformer:
         self._transformer = NodeTransformer(
             numeric_node_ids=method is None,
             aligned_node_mapping=aligned_node_mapping,
-            support_mirror_strategy=support_mirror_strategy
+            support_mirrored_strategy=support_mirrored_strategy
         )
         self._method_name = method
         self._method = EdgeTransformer.methods[self._method_name]
